import keras.ops as kops
from keras.layers import Layer, Dropout, Dense, Softmax


class MultiHeadSelfAttention(Layer):
    def __init__(
        self,
        num_heads: int = 2,
        embedding_dim: int = 64,
        projection_dim: int = None,
        qkv_bias: bool = True,
        attention_drop: float = 0.2,
        linear_drop: float = 0.2,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.embedding_dim = embedding_dim
        self.projection_dim = projection_dim if projection_dim else embedding_dim // num_heads
        self.qkv_bias = qkv_bias
        self.scale = self.projection_dim**-0.5

        self.qkv = Dense(3 * self.num_heads * self.projection_dim, use_bias=qkv_bias)
        self.proj = Dense(embedding_dim, use_bias=qkv_bias)
        self.attn_dropout = Dropout(attention_drop)
        self.linear_dropout = Dropout(linear_drop)
        self.softmax = Softmax()

    def build(self, input_shape):
        # You can perform setup tasks that depend on the input shape here
        super().build(input_shape)

    def split_heads(self, x, batch_size):
        # Split the last dimension into (num_heads, projection_dim)
        x = kops.reshape(x, (batch_size, -1, self.num_heads, self.projection_dim))
        # Transpose to shape (batch_size, num_heads, seq_len, projection_dim)
        return kops.transpose(x, axes=(0, 2, 1, 3))

    def call(self, x):
        batch_size = kops.shape(x)[0]

        # Project and reshape to (batch_size, seq_len, 3, num_heads, projection_dim)
        qkv = self.qkv(x)
        qkv = kops.reshape(qkv, (batch_size, -1, 3, self.num_heads, self.projection_dim))
        qkv = kops.transpose(qkv, axes=(0, 2, 1, 3, 4))
        q, k, v = qkv[:, 0], qkv[:, 1], qkv[:, 2]

        q = self.split_heads(q, batch_size)
        k = self.split_heads(k, batch_size)
        v = self.split_heads(v, batch_size)

        q *= self.scale

        # Attention mechanism
        attn_logits = kops.matmul(q, kops.transpose(k, axes=(0, 1, 3, 2)))
        attn = self.softmax(attn_logits)
        attn = self.attn_dropout(attn)

        weighted_avg = kops.matmul(attn, v)
        weighted_avg = kops.transpose(weighted_avg, axes=(0, 2, 1, 3))
        weighted_avg = kops.reshape(weighted_avg, (batch_size, -1, self.num_heads * self.projection_dim))

        # Output projection
        output = self.proj(weighted_avg)
        output = self.linear_dropout(output)

        return output



from typing import Optional
from dataclasses import dataclass


@dataclass
class Config_MobileViT_v1_S:
    block_1_1_dims: int = 16
    block_1_2_dims: int = 32

    block_2_1_dims: int = 64
    block_2_2_dims: int = 64
    block_2_3_dims: int = 64

    block_3_1_dims: int = 96
    block_3_2_dims: int = 96

    block_4_1_dims: int = 128
    block_4_2_dims: int = 128

    block_5_1_dims: int = 160
    block_5_2_dims: int = 160

    final_conv_dims: int = 640

    tf_block_3_dims: int = 144
    tf_block_4_dims: int = 192
    tf_block_5_dims: int = 240

    tf_block_3_repeats: int = 2
    tf_block_4_repeats: int = 4
    tf_block_5_repeats: int = 3

    depthwise_expansion_factor: int = 4


@dataclass
class Config_MobileViT_v1_XS:
    block_1_1_dims: int = 16
    block_1_2_dims: int = 32

    block_2_1_dims: int = 48
    block_2_2_dims: int = 48
    block_2_3_dims: int = 48

    block_3_1_dims: int = 64
    block_3_2_dims: int = 64

    block_4_1_dims: int = 80
    block_4_2_dims: int = 80

    block_5_1_dims: int = 96
    block_5_2_dims: int = 96

    final_conv_dims: int = 384

    tf_block_3_dims: int = 96
    tf_block_4_dims: int = 120
    tf_block_5_dims: int = 144

    tf_block_3_repeats: int = 2
    tf_block_4_repeats: int = 4
    tf_block_5_repeats: int = 3

    depthwise_expansion_factor: int = 4


@dataclass
class Config_MobileViT_v1_XXS:
    block_1_1_dims: int = 16
    block_1_2_dims: int = 16

    block_2_1_dims: int = 24
    block_2_2_dims: int = 24
    block_2_3_dims: int = 24

    block_3_1_dims: int = 48
    block_3_2_dims: int = 48

    block_4_1_dims: int = 64
    block_4_2_dims: int = 64

    block_5_1_dims: int = 80
    block_5_2_dims: int = 80

    final_conv_dims: int = 320

    tf_block_3_dims: int = 64
    tf_block_4_dims: int = 80
    tf_block_5_dims: int = 96

    tf_block_3_repeats: int = 2
    tf_block_4_repeats: int = 4
    tf_block_5_repeats: int = 3

    depthwise_expansion_factor: int = 2


def get_mobile_vit_v1_configs(model_type: str = "S", updates: Optional[dict] = None):
    if model_type == "S":
        base_config = Config_MobileViT_v1_S
    elif model_type == "XS":
        base_config = Config_MobileViT_v1_XS
    else:
        base_config = Config_MobileViT_v1_XXS

    if updates:
        return base_config(**updates)

    return base_config



from keras_vision.MobileViT_v1.base_layers import ConvLayer, InvertedResidualBlock
from keras_vision.MobileViT_v1 import multihead_self_attention_2D
from keras_vision.MobileViT_v1.mobile_vit_v1_block import Transformer, MobileViT_v1_Block

from keras_vision.MobileViT_v1.configs import (
    Config_MobileViT_v1_S,
    Config_MobileViT_v1_XS,
    Config_MobileViT_v1_XXS,
    get_mobile_vit_v1_configs,
)
from keras_vision.MobileViT_v1.mobile_vit_v1 import MobileViT_v1, build_MobileViT_v1
